function [x,histout,fail,infos] = LRGeomCG(prob, opts, x0)
% LRGEOMCG    Low-rank matrix completion by Geometric CG. 
%   Uses Armijo rule, polynomial linesearch on embedded submanifold of
%   fixed rank matrices.
%           
% Input: prob     = problem instance, see MAKE_PROB.
%        opts     = options, see DEFAULT_OPTS.           
%        x0       = starting guess.
%
% Output: x       = solution.
%         histout = iteration history. Each row of histout is
%                   [rel_norm(grad), rel_err_on_omega, relative_change, ...
%                        number of step length reductions, restarts]
%         fail    = flag for failure
%
% See also SMALL_EXAMPLE, MAKE_PROB, DEFAULT_OPTS. 

% (C) Bart Vandereycken, 2011-2012.
% Adapted from steep by C. T. Kelley, Dec 20, 1996.
%
% Paper: Low-rank matrix completion by Riemannian optimization.
%
% Info and bugs: bart.vandereycken@epfl.ch


TOL_GRAD_RELATED = 1e-3;

t_begin = tic();

if ~isfield(opts, 'beta_type')
    opts.beta_type = 'P-R';
end
if ~isfield(opts, 'linearized_step_guess')
    opts.linearized_step_guess = true;
end
if ~isfield(opts, 'sigma_armijo')
    opts.sigma_armijo = 1e-4;
end

if ~isfield(opts, 'orth_value')
    opts.orth_value = 0.1;
end

x0 = prepx(prob, x0);

beta_type = opts.beta_type;

fail = true;

norm_M_Omega = norm(prob.data);

ORTH_VALUE = opts.orth_value; % the search directions should be almost orthogonal

itc = 1; xc = x0;
fc = F(prob,xc);
gc = grad(prob,xc);
ip_gc = ip(xc,gc,gc);
rel_grad = sqrt(ip_gc)/max(1,norm(xc.sigma));
reschg = -1;
fold = 2*fc;
% first search-dir is steepest gradient
dir = scaleTxM(gc,-1);

ithist=zeros(opts.maxiter,8);
beta = 0;

objval = 2*fc / prob.m;
if opts.verbosity > 0
  fprintf('[%.4d] Cost is %5.3e \n', 0, objval);
end

% If interested in computing recovery
if opts.compute_predictions,
    preds_test = partXY(diag(xc.sigma)*xc.U',xc.V',prob.data_ts.rows,prob.data_ts.cols, prob.data_ts.nentries)';
    errors_test = preds_test - prob.data_ts.entries;
    cost_test0 = (errors_test'*errors_test)/prob.data_ts.nentries;

end

max_step = 1;


for itc=1:opts.maxiter
  if ~opts.linearized_step_guess && itc>1
    if iarm == 0,
      max_step = 2*max_step;
    elseif iarm > 1,
      max_step = 2*tinit;
    end
    tinit = max_step;
    
  else
    tinit = exact_search_onlyTxM(prob, xc,dir);

  end
   
  fc_b = fc;
  
  if ~opts.linearized_step_guess
      [xc_new,fc,succ,~,iarm, lambda, lamc] = armijo_search_mod(prob, xc,fc,gc,dir, tinit, opts.sigma_armijo); 
      max_step = lamc; % Initial stepsize guess
      tinit = lambda; % Armijo optimal stepsize
  else
      [xc_new,fc,succ,~,iarm,lambda] = armijo_search(prob, xc,fc,gc,dir, tinit, opts.sigma_armijo);
  end
  
%   if ~succ && beta ~= 0
%     if opts.verbosity > 0; disp('Line search failed on CG. Resetting to gradient.'); end
%     beta = 0;
%     % if we did cg, reset to steepest descent and try again
%     dir = scaleTxM(gc,-1);
%     tinit(itc) = exact_search_onlyTxM(prob, xc,dir);
%     [xc_new,fc,succ,~,iarm2,lambda] = armijo_search(prob, xc,fc_b,gc,dir, tinit, opts.sigma_armijo);
%     iarm = iarm2 + iarm;
%   end
  
  % if it still fails (beta is always 0 here -> steepest descent)
  % then we give up
  if ~succ
    x = xc_new;
    ithist(itc,1) = rel_grad;
    ithist(itc,2) = sqrt(2*fc)/norm_M_Omega;
    ithist(itc,3) = reschg;
    ithist(itc,4) = iarm;    
    histout=ithist(1:itc,:); 

    % the info field like with the solvers for quotient geometries
    infos.costs = histout(:,5);
    infos.all_times = histout(:,6);
    infos.beta = histout(:, 7);

    infos.costs = [ 2*F(prob,x0)/prob.m; infos.costs ];
    infos.iter_time = [ 0; infos.all_times ];
    
    if opts.compute_predictions,
        infos.test_error = histout(:,8);
        infos.test_error = [cost_test0; infos.test_error];
    end

    if opts.verbosity > 0; disp('Line search failed on steepest descent. Exiting...'); end
    return
  end
  

  
  % grad(new x)
  gc_new = grad(prob,xc_new);
  ip_gc_new = ip(xc_new,gc_new,gc_new);
  
  objval = (2*fc) / prob.m;
  % If interested in computing recovery
  if opts.compute_predictions,
    preds_test = partXY(diag(xc.sigma)*xc.U',xc.V',prob.data_ts.rows,prob.data_ts.cols, prob.data_ts.nentries)';
    errors_test = preds_test - prob.data_ts.entries;
    cost_test = (errors_test'*errors_test)/prob.data_ts.nentries;
    ithist(itc,8) = cost_test;

  end
  
  if opts.verbosity > 0
    fprintf('[%.4d] Cost is %5.3e, #extra linesearch = %i, Gradient norm sq. = %5.3e, Step = %5.3e \n', itc, objval, iarm, (ip_gc_new), lambda);
  end
  
  itc_time = toc(t_begin);
  % Test for convergence
  if objval < opts.abs_f_tol
    if opts.verbosity > 0
      disp('Abs f tol reached.')
    end
    fail = false;
    break;
  end

%   if sqrt(2*fc) < opts.abs_f_tol
%     if opts.verbosity > 0
%       disp('Abs f tol reached.')
%     end
%     fail = false;
%     break;
%   end
%   if sqrt(2*fc)/norm_M_Omega < opts.rel_f_tol
%     if opts.verbosity > 0; disp('Relative f tol reached.'); end
%     fail = false;
%     break;
%   end
  
  rel_grad = sqrt(ip_gc)/max(1,norm(xc_new.sigma));
%   if rel_grad < opts.rel_grad_tol
%     if opts.verbosity > 0; disp('Relative gradient tol reached.'); end
%     fail = false;
%     break;
%   end
  
  % for 'stagnation stopping criterion' after 10 iters
  reschg = abs(1-sqrt(2*fc)/sqrt(2*fold) );  % LMARank's detection
  %reschg = abs(sqrt(2*fc) - sqrt(2*fold)) / max(1,norm_M_Omega);
%   if itc > 10 && reschg < opts.rel_tol_change_res
%     if opts.verbosity > 0; disp('Iteration stagnated rel_tol_change_res.'); end
%     fail = true;
%     break;
%   end
  
  
%   if itc > 10
%     R1 = triu(qr([xc_new.U*diag(xc_new.sigma) xc.U*diag(xc.sigma)],0));
%     R2 = triu(qr([xc_new.V -xc.V],0));
%     
%     %rel_change_x = norm(xc_new.U*diag(xc_new.sigma)*xc_new.V' - xc.U*diag(xc.sigma)*xc.V', 'fro') / ...
%     %  norm(xc_new.U*diag(xc_new.sigma)*xc_new.V', 'fro');
%     rel_change_x = norm(R1*R2','fro')/norm(xc_new.sigma,2);
% 
%     if rel_change_x < opts.rel_tol_change_x
%       if opts.verbosity > 0; disp('Iteration stagnated for rel_tol_change_x.'); end
%       fail = true;
%       break;
%     end
%   end
%     
  % new dir = -grad(new x)
  %           + beta * vectTransp(old x, old dir, tmin*old dir)
  
  if strcmpi(beta_type, 'off')
    beta = 0;
    dir = scaleTxM(gc_new, -1);
  
  else
    
    gc_old = transpVect(prob,xc,gc,xc_new,1);  

    % we test how orthogonal the previous gradient is with
    % the current gradient, and possibly reset the to the gradient
    ip_gc_old_new = ip(xc_new,gc_old,gc_new);
    orth_grads = ip_gc_old_new / ip_gc_new;

    if abs(orth_grads) >= ORTH_VALUE 
      if opts.verbosity > 1; disp('New gradient is almost orthogonal to current gradient. This is good, so we can reset to steepest descent.'); end
      beta = 0;
      dir = plusTxM(gc_new, dir, -1, beta);

    else % Compute the CG modification
      dir = transpVect(prob,xc,dir,xc_new,1);

      % beta
      if strcmp(beta_type, 'F-R')  % Fletcher-Reeves
          beta = ip_gc_new / ip_gc;
          % new dir
          dir = plusTxM(gc_new, dir, -1, beta);
      elseif strcmp(beta_type, 'P-R')  % Polak-Ribiere+
          % vector grad(new) - transported grad(current)
          beta = (ip_gc_new - ip_gc_old_new) / ip_gc;
          beta = max(0,beta);
          dir = plusTxM(gc_new, dir, -1, beta);
          
      elseif strcmp(beta_type, 'H-S')  % Hestenes-Stiefel+
          diff = plusTxM(gc_new, gc_old, 1, -1);
          beta = ip(xc_new,diff,gc_new) / ip(xc_new,diff,dir);
          beta = max(0,beta);
          dir = plusTxM(gc_new, dir, -1, beta);
      else
          error(['Unknown params.beta_type. ' ...
              'Should be off, F-R, P-R or H-S.']);
      end
    end

    % check if dir is descent, if not take -gradient (i.e. beta=0)
    g0 = ip(xc_new,gc_new,dir);
    if g0>=0
      if opts.verbosity > 1; 
        disp('New search direction not a descent direction. Resetting to -grad.');       
      end
      dir = scaleTxM(gc_new, -1);
      beta = 0;
    else    
      ip_dir = ip(xc_new,dir,dir);
      % check if the step is sufficiently gradient-related, if not -gradient (i.e. beta=0)
      grad_related = abs(g0) / sqrt(ip_gc_new*ip_dir);

      if grad_related < TOL_GRAD_RELATED
        %if opts.verbosity > 1; 
          grad_related;
          disp('New search direction is not sufficiently gradient related. Resetting to -grad.');       
        %end
        dir = scaleTxM(gc_new, -1);
        beta = 0;
      end    
    end 
  
  end
  
  
  % update _new to current
  gc = gc_new;
  ip_gc = ip_gc_new;
  xc = xc_new;
  fold = fc;
    
  ithist(itc,1) = rel_grad;
  ithist(itc,2) = sqrt(2*fc)/norm_M_Omega;
  ithist(itc,3) = reschg;
  ithist(itc,4) = iarm; 
  ithist(itc,5) = objval;
  ithist(itc,6) = itc_time;
  ithist(itc,7) = beta;
end

x = xc_new;
ithist(itc,1) = rel_grad;
ithist(itc,2) = sqrt(2*fc)/norm_M_Omega;
ithist(itc,3) = reschg;
ithist(itc,4) = iarm; 
ithist(itc,5) = objval;
ithist(itc,6) = itc_time;
ithist(itc,7) = beta;

histout=ithist(1:itc,:); 

% the info field like with the solvers for quotient geometries
infos.costs = histout(:,5);

infos.all_times = histout(:,6);
infos.beta = histout(:, 7);

infos.costs = [ 2*F(prob,x0)/prob.m; infos.costs ];
infos.iter_time = [ 0; infos.all_times ];
if opts.compute_predictions,
    infos.test_error = histout(:,8);
    infos.test_error = [cost_test0; infos.test_error];
end

return;