function[xmin] = guess_linesearch_ubv(model, errors, searchDir, data_ls)
% [xmin, cost_new, errors_new] = guess_linesearch_ubv(model, errors,
% searchDir, data_ls)
%
%
% Computes the optimal stepsize in the direction "searchDir" at the iterate
% "model".
%
% Parameters:
% -----------
% model: Structured array with current low-rank factors
% errors: Array of the discrepancies between given entries and the predicted
%         values at the current iterate
% searchDir: Sturctured array with the search directions
% data_ls: Structured array related to given entries
%
% Output:
% -------
% xmin: Optimal stepsize

% Note:
% -----
% We find the minimum of a degree 6 polynomial
%
% This is an algorithmic implementation of
% "Low-rank optimization with trace norm penalty"
% B. Mishra, G. Meyer, F. Bach and R. Sepulchre
% Technical report, 2011
% http://arxiv.org/abs/1112.2318
%
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012



 n = data_ls.nentries;
 searchDir.U = searchDir.U; 
 searchDir.V = searchDir.V;
 searchDir.B = searchDir.B;

 term1 = (searchDir.U*model.B + model.U*searchDir.B);
 term2 = (searchDir.U*searchDir.B);

% s_star = proj_omega_sub([model.U*model.B  term1], [searchDir.V  model.V], data_ls.rows, data_ls.cols);
s_star = partXY([model.U*model.B  term1]', [searchDir.V  model.V]', data_ls.rows, data_ls.cols, n)';


% s_star_1 = proj_omega_sub([term1 term2 ], [searchDir.V  model.V ], data_ls.rows, data_ls.cols);
s_star_1 = partXY([term1 term2]', [searchDir.V  model.V ]', data_ls.rows, data_ls.cols, n)';


% s_star_ubv = proj_omega_sub(term2, searchDir.V, data_ls.rows, data_ls.cols);
s_star_ubv = partXY(term2', searchDir.V', data_ls.rows, data_ls.cols, n)';



% A degree 6 polynomial
a_6 = (s_star_ubv'*s_star_ubv); % sum(s_star_ubv.^2);
a_5 = 2*(s_star_ubv'*s_star_1);
a_4 = 2*(s_star'*s_star_ubv) + (s_star_1'*s_star_1); % 2*sum(s_star .* s_star_ubv) + sum(s_star_1.^2);
a_3 = 2*(s_star'*s_star_1) + 2*(errors'*s_star_ubv); % 2*sum(s_star .* s_star_1) + 2*sum(errors .* s_star_ubv);
a_2 = 2*(s_star_1'*errors) + (s_star'*s_star);% 2*sum(s_star_1 .* errors) + sum(s_star .^2);
a_1 = 2*(s_star'*errors); % 2*sum(s_star .* errors);
a_0 = (errors'*errors);% sum(errors.^2);

% old_cost = a_0 / n; % old cost


% Solving the derivative of the degree 6 polynomial to find the global
% minimum

coefficients = [6*a_6 , 5*a_5, 4*a_4, 3*a_3, 2*a_2, a_1];
xs = roots(coefficients);
real_roots = [];

for ii = 1 : 5
    if isreal(xs(ii)) %&& xs(ii) > 0
        real_roots = [real_roots; xs(ii)];
    end
end 

real_roots = 0.5*(real_roots + abs(real_roots));

fun_val = [];
for jj = 1 : length(real_roots)
      x = real_roots(jj);
      fun_val = [fun_val ; a_6*x^6 + a_5*x^5 + a_4*x^4 + a_3*x^3 + a_2*x^2 + a_1*x^1 + a_0];
end

[~, idx] = min(fun_val);

xmin = real_roots(idx);  % Global minimum

return;






