function[xmin] = guess_linesearch_ubv_accel_1(model, errors, searchDir, data_ls)
% [xmin, cost_new, errors_new] = guess_linesearch_ubv(model, errors,
% searchDir, data_ls)
%
%
% Computes the optimal stepsize in the direction "searchDir" at the iterate
% "model".
%
% Parameters:
% -----------
% model: Structured array with current low-rank factors
% errors: Array of the discrepancies between given entries and the predicted
%         values at the current iterate
% searchDir: Sturctured array with the search directions
% data_ls: Structured array related to given entries
%
% Output:
% -------
% xmin: Optimal stepsize

% Note:
% -----
% We find the minimum of a degree 4 polynomial
%
% This is an algorithmic implementation of
% "Low-rank optimization with trace norm penalty"
% B. Mishra, G. Meyer, F. Bach and R. Sepulchre
% Technical report, 2011
% http://arxiv.org/abs/1112.2318
%
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012



 n = data_ls.nentries;
 kappa = 1e-1;% scaling for making approximations work
 
 searchDir.U = kappa*searchDir.U; 
 searchDir.B = kappa*searchDir.B;
 searchDir.V = kappa*searchDir.V;

 
 % Accelerating by reformulation as a two factor factorization degree 6 to
 % degree 4 reduction
 
 G = model.U*model.B;
 searchDir.G = model.U*searchDir.B + searchDir.U*model.B; % Based on the assumption that the stepsize is < 1 
 
 
 s_star = partXY([G searchDir.G]', [searchDir.V model.V]', data_ls.rows, data_ls.cols, n)';


s_star_gv = partXY(searchDir.G', searchDir.V', data_ls.rows, data_ls.cols, n)';



% A degree 4 polynomial
p = s_star_gv'*s_star_gv;%   sum(s_star_gv.^2);
q = 2*(s_star_gv'*s_star); %  2*sum(s_star_gv .* s_star);
r = 2*(s_star_gv'*errors) + (s_star'*s_star);%     2*sum(s_star_gv .* errors) + sum(s_star.^2);
s = 2*sum(errors'*s_star); %2*sum(errors .* s_star);
t = (errors'*errors); % sum(errors.^2);
% old_cost = t / n % old cost


% Solving the derivative of the degree 4 polynomial to find the global
% minimum

coefficients = [4*p , 3*q, 2*r, s];
xs = roots(coefficients);
real_roots = [];


for ii = 1 : 3
    if isreal(xs(ii))% && xs(ii) > 0
        real_roots = [real_roots; xs(ii)];
    end
end 

real_roots = 0.5*(real_roots + abs(real_roots));


fun_val = [];
for jj = 1 : length(real_roots)
      x = real_roots(jj);
      fun_val = [fun_val ; p*x^4 + q*x^3 + r*x^2 + s*x + t ];
end

[~, idx] = min(fun_val);

xmin = real_roots(idx);  % Global minimum

xmin = kappa*xmin; % rescaling

return;

