function [model, infos] = CG_right_inv_uy(data_ts, data_ls, model, params)
%[model, infos] = CG_right_inv_uy(data_ts, data_ls, model, params)
%
%
% Parementers
% -----------
% model: Stuctured array with low-rank factors
%
% data_ls: data_ls.rows: rows of the given entries
%        : data_ls.cols: columns of the given entries
%        : data_ls.entries: values of the given entries
%        : data_ls.nentries: number of given entries
%
% data_ts: data_ts.Gs: Original left low-rank factor
%        : datat_ts.Hs: Original right low-rank factor
%
% params.tol : Tolerance value used in the stopping criterion
% params.verb: Verbosity
% params.sigma_armijo: The constant factor used in the Armijo linesearch
% params.ls_maxiter: Maximum of linesearch allowed
% params.beta_type : 'off' means Gradient descnet, 
%                    'P-R' means Conjugate gradient with Polak-Ribiere+,
%                    'F-R' means Conjugate gradient with Fletcher-Reeves
%                    'H-S' means Conjugate gradient with Hestenes-Stiefel+
% params.exact_linesearch : true for exact linesearch,
%                         : false for adaptive stepsize (arXiv:1209.0430, Equation (13))
% params.compute_predictions: To compute test predictions at each iteration
%                             by randomly choosing a subset of entries  
% 
% Output:
% -------
% model: Structured array with final low-rank factors
% infos: Structured array with additional information
%
%
% This is an algorithmic implementation of
% "Linear regression under fixed-rank constraints: a Riemannian approach"
% G. Meyer, S. Bonnabel and R. Sepulchre
% ICML, 2011
% and
% "Fixed-rank matrix factorizations and Riemannian low-rank optimization",
% B. Mishra, G. Meyer, S. Bonnabel and R. Sepulchre, arXiv:1209.0430
% 
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012

% Set default parameters
t_begin = tic();

params = default_params(params);
tol = params.tol;
vtol = params.vtol;
verb = params.verb;
maxiter = params.maxiter;
max_step = params.max_step;
ls_maxiter = params.ls_maxiter;

if ~isfield(params,'guess_linesearch'),
    guess_linesearch = @guess_linesearch_uy; % Non-accelerated linesearch by linearization
else
    guess_linesearch = params.guess_linesearch;% User-specified linesearch
end
fprintf('Using linesearch for UY: %s\n\n',func2str(guess_linesearch));


% Problem dimensions
ORTH_VALUE = params.orth_value; % Measure of orthogoanlity used in the CG schemes
r = size(model.U, 2);
d1 = size(model.U, 1);
d2 = size(model.Y, 1);


M_org = model.M;

% Compute cost function
n = data_ls.nentries;
preds_ls = partXY(model.U', model.Y', data_ls.rows, data_ls.cols, n)';
errors = (preds_ls - data_ls.entries);
M = model.M;
updateSparse(M, errors);
loss = (errors'*errors)/2;%   0.5*(errors.^2);

% If interested in computing recovery
if params.compute_predictions,
    preds_test = partXY(model.U', model.Y',data_ts.rows,data_ts.cols, data_ts.nentries)';
    errors_test = preds_test - data_ts.entries;
    cost_test = (errors_test'*errors_test)/data_ts.nentries;
    infos.test_error =  cost_test;
    
end

cost = loss;
grad.U = M*model.Y;
grad.U = proj_tangent_space_stiefel(grad.U, model.U);
grad.Y = M'*model.U;
grad.Y = grad.Y * model.YtY;

ip_grad_grad = inner_product_uy(model, grad, grad);

if verb,
    fprintf('[%0.4d] Cost = %5.3e\n', 0, (2/n)*cost);
end


infos.sqnormgrad = [];
infos.beta = 0;
infos.linesearch = [];
infos.costs = (2/n)*cost;
infos.iter_time = 0;
j = 0;

dir.U = -grad.U;
dir.Y = -grad.Y;
ip_grad_dir = -ip_grad_grad;
linesearch_fail = 0;

for iter = 1:maxiter,    
    % Compute first step.
    %   If we don't do an exact line search all the time,
    %   we still need it once for the starting value.
    if params.exact_linesearch || iter==1
        if iter < 5, % First few iterations, always do a non-accelerated linesearch
            alpha  = guess_linesearch_uy(model, errors, dir, data_ls); % NOT an exact search! because we use first-order approx.of expo-map on PD cone
        else
            alpha  = feval(guess_linesearch, model, errors, dir, data_ls); % NOT an exact search! because we use first-order approx.of expo-map on PD cone
        end
        
        if iter == 5 && ~strcmp(func2str(guess_linesearch),'guess_linesearch_uy'),
            fprintf('Shifted to accelerated linesearch at iteration %i...\n',iter);
        end;        % No sign correction needed, guess_linesearch gives the corrected sign
        max_step = alpha;
        model_new = model;
        
        sigma_armijo = 1e-4; % Keep a low value or 0 for an exact search
    else % Adaptive stepsize
        if j == 0,
            max_step = 2*max_step;
        elseif j > 1,
            max_step = 2*alpha;
        end
        alpha = max_step;
        alpha = -sign(ip_grad_dir)*abs(alpha); % Sign correction
        
        model_new = model;
        
        sigma_armijo = params.sigma_armijo; % User specified value
    end
    Ui = model.U;
    Yi = model.Y;
    
    model_new.U = uf(Ui + alpha * dir.U);
    model_new.Y = Yi + alpha * dir.Y;
    
    preds_ls = partXY(model_new.U', model_new.Y', data_ls.rows, data_ls.cols, n)';
    errors = (preds_ls - data_ls.entries);
    cost_new = (errors'*errors)/2;
    
    armijo = (cost - cost_new) >= sigma_armijo * abs(alpha * ip_grad_dir);
    j = 0;
    while ~armijo,
        j = j + 1;
        alpha = alpha / 2;
        model_new.U = uf(Ui + alpha * dir.U);
        model_new.Y = Yi + alpha * dir.Y;
    
        preds_ls = partXY(model_new.U', model_new.Y', data_ls.rows, data_ls.cols, n)';
        
        errors = (preds_ls - data_ls.entries);
        cost_new = (errors'*errors)/2;
        armijo = (cost - cost_new) >= sigma_armijo * abs(alpha * ip_grad_dir);
        if j > ls_maxiter,
            fprintf('*** Linesearch failed. Quitting...\n ');
            linesearch_fail = 1;
            break;
        end
    end
    
    
    model_new.YtY = model_new.Y'*model_new.Y;
    
    updateSparse(M, errors);
    
    grad_new.U = M*model_new.Y;
    grad_new.U = proj_tangent_space_stiefel(grad_new.U, model_new.U);
    grad_new.Y = M'*model_new.U;
    grad_new.Y = grad_new.Y * model_new.YtY;
    
    ip_grad_grad_new = inner_product_uy(model_new, grad_new, grad_new);
    
    infos.sqnormgrad = [infos.sqnormgrad ; ip_grad_grad_new];
    infos.linesearch = [infos.linesearch ; j];
    infos.costs = [infos.costs ; (2/n)*cost_new];
    infos.iter_time = [infos.iter_time ; toc(t_begin)];
    
    % If interested in computing recovery
    if params.compute_predictions,
        preds_test = partXY(model_new.U', model_new.Y',data_ts.rows,data_ts.cols, data_ts.nentries)';
        errors_test = preds_test - data_ts.entries;
        cost_test = (errors_test'*errors_test)/data_ts.nentries;
        infos.test_error =[infos.test_error; cost_test];
    end
    
    
    if verb,
        fprintf('[%0.4d] Cost = %7.3e, #extra linesearch = %i, Gradient norm sq. = %7.3e, Step = %7.3e \n', iter, (2/n)*cost_new, j, ip_grad_grad_new, alpha);
    end
    if (2/n)*cost_new < tol || linesearch_fail,
        break;
    end
    
    if strcmpi(params.beta_type, 'off') % Gradient Descent
        beta = 0;
        dir.U = -grad_new.U;
        dir.Y = -grad_new.Y;
        
    else
        grad_old = vector_transport_uy(model, grad, model_new); % vector transport to the new point
        orth_grads = inner_product_uy(model_new, grad_old, grad_new)/ip_grad_grad_new;
        
        if abs(orth_grads) >= ORTH_VALUE
            beta = 0;
            dir.U = -grad_new.U;
            dir.Y = -grad_new.Y;
            
        else % Compute the CG modification
            dir = vector_transport_uy(model, dir, model_new); % vector transport to the new point
            
            if strcmp(params.beta_type, 'F-R')  % Fletcher-Reeves
                % vector grad(new) - transported grad(current)
                beta = ip_grad_grad_new / ip_grad_grad;
                
            elseif strcmp(params.beta_type, 'P-R')  % Polak-Ribiere+
                % vector grad(new) - transported grad(current)
                diff.U = grad_new.U - grad_old.U;
                diff.Y = grad_new.Y - grad_old.Y;
                ip_diff = inner_product_uy(model_new, grad_new, diff);
                beta = ip_diff / ip_grad_grad;
                beta = max(0, beta);
           
            elseif strcmp(params.beta_type, 'H-S')  % Hestenes-Stiefel+
                diff.U = grad_new.U - grad_old.U;
                diff.Y = grad_new.Y - grad_old.Y;
                ip_diff = inner_product_uy(model_new, grad_new, diff);
                ip_diff_dir = inner_product_uy(model_new, diff, dir);
                beta = ip_diff / ip_diff_dir;
                beta = max(0, beta);
            else
                error(['Unknown params.beta_type. ' ...
                    'Should be off, F-R, P-R or H-S.']);
            end
            
            dir.U = -grad_new.U + beta*dir.U;
            dir.Y = -grad_new.Y + beta*dir.Y;
            
        end
        
    end
    ip_grad_dir_new = inner_product_uy(model_new, dir, grad_new);
    
    % If CG update produces an accent direction, we either reverse it or
    % switch the negative gradient direction.
    if sign(ip_grad_dir_new) > 0,
        %         % Either we reverse it.
        %         ip_grad_dir_new = -ip_grad_dir_new;
        %         dir.U = -dir.U;
        %         dir.Y = -dir.Y;
        
        %         if verb,
        %             fprintf('CG warning: An accent direction obtained. Reversed it.\n');
        %         end
        
        % Or we switch to the negative gradient direction.
        dir.U = -grad_new.U;
        dir.Y = -grad_new.Y;
        beta = 0;
        ip_grad_dir_new = -ip_grad_grad_new;
        if verb,
            fprintf('CG warning: An accent direction obtained. Switched to negative gradient dir.\n');
        end
        
    end
    
    % Note: By this time we should have properly chosen the CG direction.
    % Next: We collect the information and prepare for the next iteration.
    
    
    
    infos.beta = [infos.beta ; beta];
    
    grad = grad_new;
    ip_grad_grad = ip_grad_grad_new;
    ip_grad_dir = ip_grad_dir_new;
    model = model_new;
    cost = cost_new;
    
    
    
end

model = model_new;

return;



