function[xmin] = guess_linesearch_uy_accel(model, errors, searchDir, data_ls)
% [xmin, cost_new, errors_new] = guess_linesearch_uy_accel(model, errors,
% searchDir, data_ls)
%
%
% Computes the optimal stepsize in the direction "searchDir" at the iterate
% "model".
%
% Parameters:
% -----------
% model: Structured array with current low-rank factors
% errors: Array of the discrepancies between given entries and the predicted
%         values at the current iterate
% searchDir: Sturctured array with the search directions
% data_ls: Structured array related to given entries
%
% Output:
% -------
% xmin: Optimal stepsize

% Note:
% -----
% We find the minimum of a degree 2 polynomial
%
% This is an algorithmic implementation of
% "Low-rank optimization with trace norm penalty"
% B. Mishra, G. Meyer, F. Bach and R. Sepulchre
% Technical report, 2011
% http://arxiv.org/abs/1112.2318
%
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012




 n = data_ls.nentries;
 kappa = 1e-1; % scaling for making approximations work
 searchDir.U = kappa*searchDir.U; % scaling
 searchDir.Y = kappa*searchDir.Y;

 
 % Accelerating by reformulation as one factor factorization with degree 4 to
 % degree 2 reduction
 
 
s_star = partXY([model.U searchDir.U]', [searchDir.Y model.Y]', data_ls.rows, data_ls.cols, n)';

term_num = (errors'*s_star);
term_den = (s_star'*s_star);

xmin = - term_num/term_den;
xmin = max(0, xmin);

xmin = kappa*xmin; % rescaling

return;


