% This is a generic test file for calling the algorithm
% for the low-rank matrix completion problem.
%
% Refer "A Riemannian geometry for low-rank matrix completion",
% B. Mishra, K. Adithya Apuroop and R. Sepulchre,
% Technical report (arXiv:1211.1550), 2012.
%
% Author:
% Bamdev Mishra
% b.mishra@ulg.ac.be

clear all; clc; close all;

% Set randstream based on clock
s = RandStream.create('mt19937ar','seed',sum(100*clock));
RandStream.setDefaultStream(s);


%% Problem specifications
d1 = 5000;
d2 = 5000;
true_rank = 5;
over_sampling = 5;
fprintf('Random rank %i matrix of size %i times %i and OS = %i\n', true_rank, d1, d2, over_sampling);

tol = 1e-20;
verbosity = 0; % Display information
max_iter = 500; % Maximu iterations
max_iter_TR = 100; % Maximum iterations for TR
max_inner = 100; % Maximum iterations for the inner subproblem in TR
guess_Delta = true; % Guess the initial trust-region radius using the exact/linearized linesearch

ls_maxiter = 10; % Maximum linesearch steps allowed
compute_predictions = true; % Compute test predictions every iteration
random_initialization = false; % Initialization
ORTH_VALUE = 0.1; % Measure of othogoanlity


%% Generate data
dof = (d1 + d2 - true_rank)*true_rank;
prop_known = over_sampling * dof / (d1 * d2);

m = round(dof*over_sampling);
Matcom = prob_matrix(d1,d2,true_rank,m); % Problem related structure variable,


%% Sparse structure
sparse_structure = sparse(Matcom.row', Matcom.col', Matcom.values_Omega*1, d1, d2, m); % To be used while updating sparse matrices


%% Initialization for algorithms
r = true_rank;
if ~random_initialization,
    [U B V] = svds(sparse_structure, r);
    G = U*(B.^(0.5));
    H = V*(B.^(0.5));
    fprintf('**** Initialization by taking %i dominant SVD\n', r);
else
    G = randn(d1, r); H = randn(d2, r);
    fprintf('**** Random initialization\n');
end




%% Interface with our algorithms
data_ls.rows = Matcom.row';
data_ls.cols = Matcom.col';
data_ls.entries = Matcom.values_Omega';
data_ls.nentries = length(data_ls.entries);
data_ts.Gs = Matcom.left;
data_ts.Hs = Matcom.right;

% Testing data
data_ts.nentries = data_ls.nentries;
data_ts.rows = randi(d1, data_ts.nentries, 1);
data_ts.cols = randi(d2, data_ts.nentries, 1);
data_ts.entries = partXY(data_ts.Gs', data_ts.Hs',data_ts.rows,data_ts.cols,data_ts.nentries)';

%% GH (New geometry): Proposed by Mishra et al., 2012
clear model;

model.G = G;
model.H = H;
model.d1 = d1;
model.d2 = d2;
model.GtG = G'*G;
model.HtH = H'*H;
model.M = sparse_structure; % Sparse structure

params.tol = tol;
params.max_step = []; % Not required as we do an exact search at least for the first iteration
params.maxiter = max_iter;
params.verb = verbosity;
params.compute_predictions = compute_predictions;
params.orth_value = ORTH_VALUE;

%%
fprintf('--------------- Adaptive stepsize: qGeomMC GD -------------\n');

params.beta_type = 'off';
params.exact_linesearch = false;
params.sigma_armijo = 0.5; % Armijo backtracking

t=tic();
[model_GD_gh, infos_GD_gh] = CG_gh(data_ts, data_ls, model, params);
toc(t)
fprintf('Average #extra-linesearch = %f \n',mean(infos_GD_gh.linesearch));


%%
fprintf('--------------- Exact linesearch: qGeomMC GD -------------\n');

params.beta_type = 'off';
params.exact_linesearch = true;
params.sigma_armijo = 1e-4;% Armijo backtracking

t=tic();
[model_GD_gh_exact, infos_GD_gh_exact] = CG_gh(data_ts, data_ls, model, params);
toc(t)


%%
fprintf('--------------- Adaptive stepsize: qGeomMC CG -------------\n');

params.beta_type = 'P-R';
params.exact_linesearch = false;
params.sigma_armijo = 0.5; % Armijo backtracking

t=tic();
[model_CG_gh, infos_CG_gh] = CG_gh(data_ts, data_ls, model, params);
toc(t)
fprintf('Average #extra-linesearch = %f \n',mean(infos_CG_gh.linesearch));



%%
fprintf('--------------- Exact linesearch: qGeomMC CG -------------\n');

params.beta_type = 'P-R';
params.exact_linesearch = true;
params.sigma_armijo = 1e-4; % Armijo backtracking

t=tic();
[model_CG_gh_exact, infos_CG_gh_exact] = CG_gh(data_ts, data_ls, model, params);
toc(t)




%% Interface with the Trust-region algorithm

% Sparse skeleton structure for efficient sparse update
prob.temp_omega = sparse(Matcom.row', Matcom.col', Matcom.values_Omega'*1, Matcom.size1, Matcom.size2, m);
prob.temp_omega_2 = sparse(Matcom.row', Matcom.col', Matcom.values_Omega'*2, Matcom.size1, Matcom.size2, m);

% Options for the trust-region
par.opts.tol_cost = tol;
par.rank = r;
par.maxiter = max_iter_TR;
par.max_inner = max_inner;
par.opts.verbosity = verbosity;
par.opts.compute_predictions = compute_predictions;
par.opts.guess_Delta = guess_Delta;


%
prob.r = r;
prob.k = r;
prob.n1 = Matcom.size1;
prob.n2 = Matcom.size2;
prob.m = length(Matcom.ind_Omega);

prob.Omega_i = Matcom.row';
prob.Omega_j = Matcom.col';
prob.data = Matcom.values_Omega'; % we need a column vector!
prob.Omega = Matcom.ind_Omega';


% Testing data
prob.data_ts.m = data_ts.nentries;
prob.data_ts.rows = data_ts.rows;
prob.data_ts.cols = data_ts.cols;
prob.data_ts.entries = data_ts.entries;

par.prob = prob;

fprintf('----------- qGeomMC TR ---------- \n');

clear x ;
x.G = G;
x.H = H;
x.GtG = x.G'*x.G;
x.HtH = x.H'*x.H;
x.invGtG = x.GtG \ eye(size(x.GtG));
x.invHtH = x.HtH \ eye(size(x.HtH));
x = prepx_gh(prob, x);
if ~guess_Delta,
    x.Delta0 = min(d1, d2)*r;
end
par.x = x;

[X_GH, out_GH] = geom_rtr_common_gh(Matcom,par);

fprintf('Total time is %f \n', X_GH.infos.iter_time(end));

%% plots
% Scaling for plots
X_GH.infos.costs = 2*X_GH.infos.costs/prob.m; % mse


fs = 20;
figure,
semilogy((infos_GD_gh.iter_time),(infos_GD_gh.costs),'Color','b','LineWidth',2);
hold on;
semilogy((infos_GD_gh_exact.iter_time),(infos_GD_gh_exact.costs),'Color','r','LineWidth',2);
semilogy((infos_CG_gh.iter_time),(infos_CG_gh.costs),'Color','b','LineWidth',2,'LineStyle','--');
semilogy((infos_CG_gh_exact.iter_time),(infos_CG_gh_exact.costs),'Color','r','LineWidth',2,'LineStyle','--');
semilogy(X_GH.infos.iter_time,X_GH.infos.costs,'-.x','Color','green','LineWidth',2,'markerSize',10);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Time','FontSize',fs);
ylabel(ax1,'Cost','FontSize',fs);
legend('Adaptive GD', 'Exact GD','Adaptive CG', 'Exact CG', 'Trust-region');
legend 'boxoff';
title(['qGeomMC: Adaptive, exact and TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling)])
axis([get(gca,'XLim') tol 1e1])
box off;



fs = 20;
figure,
semilogy(0:length(infos_GD_gh.costs)-1,(infos_GD_gh.costs),'Color','b','LineWidth',2);
hold on;
semilogy(0:length(infos_GD_gh_exact.costs)-1,(infos_GD_gh_exact.costs),'Color','r','LineWidth',2);
semilogy(0:length(infos_CG_gh.costs)-1,(infos_CG_gh.costs),'Color','b','LineWidth',2,'LineStyle','--');
semilogy(0:length(infos_CG_gh_exact.costs)-1,(infos_CG_gh_exact.costs),'Color','r','LineWidth',2,'LineStyle','--');
semilogy(0 : length(X_GH.infos.costs) - 1, X_GH.infos.costs,'-.x','Color','green','LineWidth',2,'markerSize',10);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Number of iterations','FontSize',fs);
ylabel(ax1,'Cost','FontSize',fs);
legend('Adaptive GD', 'Exact GD','Adaptive CG', 'Exact CG', 'Trust-region');
legend 'boxoff';
axis([get(gca,'XLim') tol 1e1])
title(['qGeomMC: Adaptive, exact and TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling)])
box off;



if compute_predictions,
    X_GH.infos.test_error = 2*X_GH.infos.test_error/prob.data_ts.m; %scaling
    
    fs = 20;
    figure,
    semilogy(0:length(infos_GD_gh.test_error)-1,(infos_GD_gh.test_error),'Color','b','LineWidth',2);
    hold on;
    semilogy(0:length(infos_GD_gh_exact.test_error)-1,(infos_GD_gh_exact.test_error),'Color','r','LineWidth',2);
    semilogy(0:length(infos_CG_gh.test_error)-1,(infos_CG_gh.test_error),'Color','b','LineWidth',2,'LineStyle','--');
    semilogy(0:length(infos_CG_gh_exact.test_error)-1,(infos_CG_gh_exact.test_error),'Color','r','LineWidth',2,'LineStyle','--');
    semilogy(0:length(X_GH.infos.test_error)-1,(X_GH.infos.test_error),'-.x','Color','green','LineWidth',2,'markerSize',10);
    
    hold off;
    ax1 = gca;
    set(ax1,'FontSize',fs);
    xlabel(ax1,'Number of iterations','FontSize',fs);
    ylabel(ax1,'Test error','FontSize',fs);
    legend('Adaptive GD', 'Exact GD','Adaptive CG', 'Exact CG', 'Trust-region');
    legend 'boxoff';
    axis([get(gca,'XLim') tol 1e1])
    title(['qGeomMC: Adaptive, exact and TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling)])
    box off;
end

