% This is a generic test file comparing fixed-rank optimization algorithms
% for the low-rank matrix completion problem.
%
% Algorithm type: Riemannian Conjugate gradient algorithm with exact or
% adaptive stepsize guess
%
% Bechmarks: LRGeom by Bart Vandereycken, http://web.math.princeton.edu/~bartv/matrix_completion.html
%            Right invariant GH^T and UY^T by B. Mishra et al., (arXiv:1209.0430)
%            LMaFit by Wen et. al, 2010
%            ScGrassMC by Ngo et. al, 2012
%            UBV^T factorization by B. Mishra et al., (arXiv:1112.2318)
%
% Refer "A Riemannian geometry for low-rank matrix completion",
% B. Mishra, K. Adithya Apuroop and R. Sepulchre,
% Technical report (arXiv:1211.1550), 2012.
%
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012

cd ..

clear all; clc; close all;

% Set randstream based on clock
s = RandStream.create('mt19937ar','seed', sum(100*clock));
RandStream.setDefaultStream(s);



%% Problem specifications
d1 = 5000;
d2 = 5000;
true_rank = 5;
over_sampling = 4;
fprintf('Random rank %i matrix of size %i times %i and OS = %i\n', true_rank, d1, d2, over_sampling);

% Algorithm specifications
beta_type = 'P-R'; % Algorithm type
exact_linesearch = true; % Linesearch type
fprintf('beta update for conjugate gradient schemes: %s \n',beta_type)
fprintf('Exact linesearch: %i \n', exact_linesearch);

% Options
tol = 1e-10;
verbosity = false;
max_iter = 250;
ls_maxiter = 50; % Maximum linesearch steps allowed
compute_predictions = false; % Compute recovery at each iteration when asked
random_initialization = true; % Initialization
ORTH_VALUE = 0.1; Inf; 0.2; % Measure of othogoanlity


sigma_armijo = 1e-4; % Armijo backtracking
if ~exact_linesearch,
    sigma_armijo = 0.5;
end


%% Generate data well-conditioned or ill-conditioned
dof = (d1 + d2 - true_rank)*true_rank;
prop_known = over_sampling * dof / (d1 * d2);

% Creating a well/ill-conditioned matrix

prob_options.type ='2factors'; % Well-conditioned
% prob_options.type ='decaying'; % Decaying singular values

if strcmp(prob_options.type,'decaying'),
    
    s1 = 1000;
    step = 1000; S0 = diag([s1:step:s1+(true_rank-1)*step]*1); % Linear decay
    %     S0 = s1*diag(logspace(-2,0,true_rank)); % Exponential decay
    
    prob_options.diag = S0;
    prob_options.scaling = 1;
    S0 = prob_options.scaling*prob_options.diag;
    
    fprintf('Creating a matrix with singular values...\n')
    for kk = 1: length(diag(S0));
        fprintf('%s \n', num2str(S0(kk, kk), '%10.5e') );
    end
    
end

m = round(prop_known*d1*d2);
Matcom = prob_matrix(d1,d2,true_rank,m,prob_options); % Problem related structure variable
singular_vals = svd(Matcom.left'*Matcom.left);
condition_number = sqrt(max(singular_vals)/min(singular_vals));
fprintf('Condition number is %f \n', condition_number);

%% Sparse structure
sparse_structure = sparse(Matcom.row', Matcom.col', Matcom.values_Omega*1, d1, d2, m); % To be used while updating sparse matrices


%% Initialization for algorithms
r = true_rank;
if ~random_initialization,
    [U B V] = svds(sparse_structure, r);
    G = U*(B.^(0.5));
    H = V*(B.^(0.5));
    fprintf('**** Initialization by taking %i dominant SVD\n', r);
else
    G = randn(d1, r); H = randn(d2, r);
    [Qg Rg] = qr(G, 0);
    [Qh Rh] = qr(H, 0);
    [q1 b q2] = svd(Rg * Rh');
    U = Qg * q1;
    V = Qh * q2;
    B = b;
    fprintf('**** Random initialization\n');
end


%% Interface with our algorithms
data_ls.rows = Matcom.row';
data_ls.cols = Matcom.col';
data_ls.entries = Matcom.values_Omega';
data_ls.nentries = length(data_ls.entries);
data_ts.Gs = Matcom.left;
data_ts.Hs = Matcom.right;

% Testing data
data_ts.nentries = data_ls.nentries;
data_ts.rows = randi(d1, data_ts.nentries, 1);
data_ts.cols = randi(d2, data_ts.nentries, 1);
data_ts.entries = partXY(data_ts.Gs', data_ts.Hs',data_ts.rows,data_ts.cols,data_ts.nentries)';


%% GH (New geometry): by Mishra et al., 2012
clear model;

model.G = G;
model.H = H;
model.d1 = d1;
model.d2 = d2;
model.GtG = G'*G;
model.HtH = H'*H;
model.M = sparse_structure; % Sparse structure

params.tol = tol;
params.max_step = []; % Not required as we do an exact search
params.maxiter = max_iter;
params.ls_maxiter = ls_maxiter;
params.sigma_armijo = sigma_armijo;
params.compute_predictions = compute_predictions;
params.beta_type = beta_type;
params.exact_linesearch = exact_linesearch;
params.verb = verbosity;
params.orth_value = ORTH_VALUE;


fprintf('--------------- qGeomMC CG -------------\n');

tic;
[model_CG_gh, infos_CG_gh] = CG_gh(data_ts, data_ls, model, params);
toc;


%% Right invariant GH: by Meyer et al., 2011
clear model;

model.G = G;
model.H = H;
model.d1 = d1;
model.d2 = d2;
model.GtG = G'*G;
model.HtH = H'*H;
model.M = sparse_structure; % Sparse structure

params.tol = tol;
params.max_step = [];% Not required as we do an exact search
params.maxiter =  max_iter;
params.ls_maxiter = ls_maxiter;
params.sigma_armijo = sigma_armijo;
params.compute_predictions = compute_predictions;
params.beta_type = beta_type;
params.exact_linesearch = exact_linesearch;
params.verb = verbosity;
params.orth_value = ORTH_VALUE;


fprintf('--------------- CG - Right invariant GH^T -------------\n');

tic;
[model_CG_right_inv_gh, infos_CG_right_inv_gh] = CG_right_inv_gh(data_ts, data_ls, model, params);
toc;



%% UBV: by Mishra et al., 2011
clear model ;

model.U = U;
model.V = V;
model.B = B;
model.d1 = d1;
model.d2 = d2;
model.M = sparse_structure; % Sparse structure

params.tol = tol;
params.max_step = []; % Not required as we do an inexact (linearized) search
params.maxiter = max_iter;
params.ls_maxiter = ls_maxiter;
params.sigma_armijo = sigma_armijo;
params.compute_predictions = compute_predictions;
params.beta_type = beta_type;
params.exact_linesearch = exact_linesearch;
params.verb = verbosity;
params.orth_value = ORTH_VALUE;
params.guess_linesearch = @guess_linesearch_ubv_accel_2;
% Note: We provide 3 linesearches,
% *_ubv > *_ubv_accel_1 > *_ubv_accel_2
% in computational burden as well as increasing order of approxmiation.
% The accelerated versions are based on the assumption that the optimal stepsize
% for this geometry is < 1. This is an empirical observation.


fprintf('--------------- UBV - CG -------------\n');
tic;
[model_CG_ubv, infos_CG_ubv] = CG_ubv(data_ts, data_ls, model, params);
toc;


%% UY: by Mishra et al., 2012
clear model params

model.U = U;
model.Y = V*B;
model.d1 = d1;
model.d2 = d2;
model.YtY = model.Y'*model.Y;
model.M = sparse_structure; % Sparse structure

params.tol = tol;
params.max_step = [];% Not required as we do an exact search
params.maxiter =  max_iter;
params.sigma_armijo = sigma_armijo; % Bart
params.compute_predictions = compute_predictions;
params.beta_type = beta_type;
params.exact_linesearch = exact_linesearch;
params.verb = verbosity;
params.orth_value = ORTH_VALUE;
params.guess_linesearch = @guess_linesearch_uy_accel;
% Note: We provide 2 linesearches,
% *_uy > *_uy_accel
% in computational burden as well as increasing order of approxmiation.
% The accelerated versions are based on the assumption that the optimal stepsize
% for this geometry is < 1. This is an empirical observation.


fprintf('--------------- UY - CG --------------\n');
tic;
[model_CG_right_inv_uy, infos_CG_right_inv_uy] = CG_right_inv_uy(data_ts, data_ls, model, params);
toc;


%% Embedded geometry: LRGeom by Vandereycken, 2012

% Initialization
prob.r = true_rank;
prob.k = true_rank;
prob.n1 = Matcom.size1;
prob.n2 = Matcom.size2;
prob.m = length(Matcom.ind_Omega);
prob.Omega_i = Matcom.row';
prob.Omega_j = Matcom.col';
prob.data = Matcom.values_Omega'; % we need a column vector!
prob.Omega = Matcom.ind_Omega';
prob.left = Matcom.left;
prob.right = Matcom.right;
prob.temp_omega = sparse_structure;
prob.data_ts = data_ts;

fprintf('-------------- LRGeom CG ------------ \n');

options = default_opts();
x0.U = U;
x0.sigma = diag(B);
x0.V = V;
prob.mu = 0;

options.abs_f_tol = tol;
options.maxiter = max_iter;
options.verbosity = verbosity;
options.compute_predictions = compute_predictions;
options.beta_type = beta_type;
options.linearized_step_guess = exact_linesearch;
options.sigma_armijo = sigma_armijo;
options.orth_value = Inf; ORTH_VALUE; % As suggested by BV



tic;
[Xcg,hist,~,infos_lrgeom_cg] = LRGeomCG(prob, options, x0);
toc;


%% Run LMaFit

Known.Ik = Matcom.row';
Known.Jk = Matcom.col';
Known.entries = Matcom.values_Omega';
data_ts.Gs = Matcom.left;
data_ts.Hs = Matcom.right;

opts.X = G;
opts.Y = H';
opts.maxit = max_iter;
opts.tol = tol;
opts.verbosity = verbosity;
opts.compute_predictions = compute_predictions;

fprintf('---------------LMaFit-------------\n');

tic;
[X_lma ,Y_lma, infos_lma] = mod_lmafit_mc_adp(d1,d2,Known, data_ts, opts);
toc;



%% ScGrassMC by Ngo and Saad, 2012
fprintf('-------------- ScGrassMC -------------- \n');
bt = tic;
data_ls.sparse_structure = sparse_structure;
exact_linesearch_sc = exact_linesearch;% the linesearch seems problematic without linesearch!!!
max_iter_sc = max_iter;
% type help ScGrassMC for options for ScGrassMC
[U_sc S_sc V_sc hist_sc] = mod_scaled_grass_mc(data_ts, data_ls, true_rank,...
    'tol', tol,...
    'maxit',max_iter_sc,...
    'grad_type','scaled',...
    'beta_type',beta_type,...
    'sigma_type','approx',... %'tol_reschg', tol_reschg,...
    'verbose', verbosity,...
    'U', U,...
    'V', V,...
    'S', B,...
    'orth_value', Inf,... % As suggested by the authors
    'compute_predictions', compute_predictions, ...
    'linearized_step_guess', exact_linesearch_sc); % linearize linesearch?

toc(bt);


%% plots

% Cost versus time
fs = 20;
figure;
semilogy((infos_CG_gh.iter_time),(infos_CG_gh.costs),'--','Color','b','LineWidth',2);
hold on;
semilogy((infos_CG_right_inv_gh.iter_time),(infos_CG_right_inv_gh.costs),'--','Color','r','LineWidth',2);
semilogy((infos_lrgeom_cg.iter_time),(infos_lrgeom_cg.costs),'--','Color','black','LineWidth',2);
semilogy((infos_lma.iter_time),(infos_lma.costs),'--','Color',[.91 .41 .17],'LineWidth',2);
semilogy((hist_sc.iter_time),(hist_sc.costs),'--','Color','m','LineWidth',2);
semilogy((infos_CG_ubv.iter_time),(infos_CG_ubv.costs),'--','Color','green','LineWidth',2);
semilogy((infos_CG_right_inv_uy.iter_time),(infos_CG_right_inv_uy.costs),'--','Color','c','LineWidth',2);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Time','FontSize',fs);
ylabel(ax1,'Cost','FontSize',fs);
axis([get(gca,'XLim') tol 1e1])
legend('qGeomMC CG','Right inv. GH^T CG','LRGeom CG','LMaFit','ScGrassMC CG', 'UBV^T CG', 'UY^T CG');
legend 'boxoff';
title([num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling),', Condition No. ', num2str(condition_number)])
box off;


% Cost versus iterations
fs = 20;
figure;
semilogy(0:length(infos_CG_gh.costs)-1,(infos_CG_gh.costs),'--','Color','b','LineWidth',2);
hold on;
semilogy(0:length(infos_CG_right_inv_gh.costs)-1,(infos_CG_right_inv_gh.costs),'--','Color','r','LineWidth',2);
semilogy(0:length(infos_lrgeom_cg.costs)-1,(infos_lrgeom_cg.costs),'--','Color','black','LineWidth',2);
semilogy(0:length(infos_lma.costs)-1,(infos_lma.costs),'--','Color',[.91 .41 .17],'LineWidth',2);
semilogy(0:length(hist_sc.costs)-1,(hist_sc.costs),'--','Color','m','LineWidth',2);
semilogy(0:length(infos_CG_ubv.costs)-1,(infos_CG_ubv.costs),'--','Color','g','LineWidth',2);
semilogy(0:length(infos_CG_right_inv_uy.costs)-1,(infos_CG_right_inv_uy.costs),'--','Color','cyan','LineWidth',2);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Number of iterations','FontSize',fs);
ylabel(ax1,'Cost ','FontSize',fs);
axis([get(gca,'XLim') tol 1e1])
legend('qGeomMC CG','Right inv. GH^T CG','LRGeom CG','LMaFit','ScGrassMC CG', 'UBV^T CG', 'UY^T CG');
legend 'boxoff';
box off;
title([num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling),', Condition No. ', num2str(condition_number)])


% beta versus iterations
fs = 20;
figure,
plot(infos_CG_gh.beta,'color','b','LineWidth',2);
hold on;
plot(infos_CG_right_inv_gh.beta,'color','r','LineWidth',2);
plot(infos_lrgeom_cg.beta,'color','black','LineWidth',2);
plot(hist_sc.beta, 'color', 'm', 'LineWidth',2);
plot(infos_CG_ubv.beta,'color','green','LineWidth',2);
plot(infos_CG_right_inv_uy.beta,'color','c','LineWidth',2);


hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Number of iterations','FontSize',fs);
ylabel(ax1,'\beta','FontSize',fs);
legend('qGeomMC CG','Right inv. GH^T CG','LRGeom CG','ScGrassMC CG','UBV^T CG', 'UY^T CG');
legend 'boxoff';
box off;


% Test error versus iterations
if compute_predictions,
    fs = 20;
    figure;
    semilogy(0:length(infos_CG_gh.test_error)-1,(infos_CG_gh.test_error),'--','Color','b','LineWidth',2);
    hold on;
    semilogy(0:length(infos_CG_right_inv_gh.test_error)-1,(infos_CG_right_inv_gh.test_error),'--','Color','r','LineWidth',2);
    semilogy(0:length(infos_lrgeom_cg.test_error)-1,(infos_lrgeom_cg.test_error),'--','Color','black','LineWidth',2);
    semilogy(0:length(infos_lma.test_error)-1,(infos_lma.test_error),'--','Color',[.91 .41 .17],'LineWidth',2);
    semilogy(0:length(hist_sc.test_error)-1,(hist_sc.test_error),'--','Color','m','LineWidth',2);
    semilogy(0:length(infos_CG_ubv.test_error)-1,(infos_CG_ubv.test_error),'--','Color','green','LineWidth',2);
    semilogy(0:length(infos_CG_right_inv_uy.test_error)-1,(infos_CG_right_inv_uy.test_error),'--','Color','c','LineWidth',2);
    
    hold off;
    ax1 = gca;
    set(ax1,'FontSize',fs);
    xlabel(ax1,'Number of iterations','FontSize',fs);
    ylabel(ax1,'Test error','FontSize',fs);
    legend('qGeomMC CG','Right inv. GH^T CG','LRGeom CG','LMaFit','ScGrassMC CG', 'UBV^T CG', 'UY^T CG');
    legend 'boxoff';
    box off;
    title([num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ',num2str(over_sampling),', Condition No. ', num2str(condition_number)])
end





