% This is a generic test file for calling the algorithm
% for the low-rank matrix completion problem.
%
% Algorithm type: Riemannian Trust-region using the generic toolbox GenRTR
%
% Bechmarks: LRGeom by Bart Vandereycken, http://web.math.princeton.edu/~bartv/matrix_completion.html
%            Right invariant GH^T and UY^T by B. Mishra et al., (arXiv:1209.0430)
%            UBV^T factorization by B. Mishra et al., (arXiv:1112.2318)
%            RTRMC by N. Boumal and P.-A. Absil, NIPS, 2011
%
% Refer "A Riemannian geometry for low-rank matrix completion",
% B. Mishra, K. Adithya Apuroop and R. Sepulchre,
% Technical report (arXiv:1211.1550), 2012.
%
% Author:
% Bamdev Mishra
% b.mishra@ulg.ac.be

cd ..

clear all; clc; close all;

% Set randstream based on clock
s = RandStream.create('mt19937ar','seed',sum(100*clock));
RandStream.setDefaultStream(s);


%% Problem specifications
d1 = 5000;
d2 = 5000;
true_rank = 5;
over_sampling = 4;
fprintf('Random rank %i matrix of size %i times %i and OS = %i\n', true_rank, d1, d2, over_sampling);

% Options
tol = 1e-20;
verbosity = 0; % Must be an integer; Display information
max_iter = 100; % Maximum iterations
max_inner = 100; % Maximum iterations for the inner subproblem
guess_Delta = true; % Guess the initial trust-region radius using the exact/linearized linesearch
compute_predictions = true; % Compute test predictions every iteration
random_initialization = false; % Initialization


%% Generating data
dof = (d1 + d2 - true_rank)*true_rank;
prop_known = over_sampling * dof / (d1 * d2);

% Creating an ill-conditioned matrix
% Change S0 to suit your requirements

prob_options.type ='2factors'; % Well-conditioned
% prob_options.type ='decaying'; % Decaying singular values

if strcmp(prob_options.type,'decaying'),
    
    s1 = 1000;
    step = 1000; S0 = diag([s1:step:s1+(true_rank-1)*step]*1); % Linear decay
    %     S0 = s1*diag(logspace(-2,0,true_rank)); % Exponential decay
    
    prob_options.diag = S0;
    prob_options.scaling = 1;
    S0 = prob_options.scaling*prob_options.diag;
    
    fprintf('Creating a matrix with singular values...\n')
    for kk = 1: length(diag(S0));
        fprintf('%s \n', num2str(S0(kk, kk), '%10.5e') );
    end
end

m = round(prop_known*d1*d2);
Matcom = prob_matrix(d1,d2,true_rank,m,prob_options);
singular_vals = svd(Matcom.left'*Matcom.left);
condition_number = sqrt(max(singular_vals)/min(singular_vals));
fprintf('Condition number is %f \n', condition_number);


% Sparse skeleton structure for efficient sparse update
prob.temp_omega = sparse(Matcom.row', Matcom.col', Matcom.values_Omega'*1, Matcom.size1, Matcom.size2, m);
prob.temp_omega_2 = sparse(Matcom.row', Matcom.col', Matcom.values_Omega'*2, Matcom.size1, Matcom.size2, m);


%% Initialization for algorithms
r = true_rank;
if ~random_initialization,
    [U0 B0 V0] = svds(prob.temp_omega, r);
    G0 = U0*(B0.^(0.5));
    H0 = V0*(B0.^(0.5));
    fprintf('**** Initialization by taking %i dominant SVD\n', r);
else
    G0 = randn(d1, r); H0 = randn(d2, r);
    fprintf('**** Random initialization\n');
    [Q1 R1] = qr(G0, 0);
    [Q2 R2] = qr(H0, 0);
    [qL S qR] = svd(R1*R2');
    U0 = Q1*qL;
    B0 = S;
    V0 = Q2*qR;
end



%% Options
par.opts.tol_cost = tol;
par.rank = r;
par.maxiter = max_iter;
par.max_inner = max_inner;
par.opts.verbosity = verbosity;
par.opts.compute_predictions = compute_predictions;
par.opts.guess_Delta = guess_Delta;


%% Interface with the Trust-region algorithm
prob.r = r;
prob.k = r;
prob.n1 = Matcom.size1;
prob.n2 = Matcom.size2;
prob.m = length(Matcom.ind_Omega);

prob.Omega_i = Matcom.row';
prob.Omega_j = Matcom.col';
prob.data = Matcom.values_Omega'; % we need a column vector!
prob.Omega = Matcom.ind_Omega';


% Testing data
prob.data_ts.m = prob.m;
prob.data_ts.rows = randi(d1, prob.data_ts.m, 1);
prob.data_ts.cols = randi(d2, prob.data_ts.m, 1);
prob.data_ts.entries = partXY(Matcom.left', Matcom.right', prob.data_ts.rows, prob.data_ts.cols, prob.data_ts.m)';



par.prob = prob;



%% LRGeom RTR by Vandereycken 2011
fprintf('----------- LRGeom RTR ---------- \n');

clear x;
x.V = V0;
x.sigma = diag(B0);
x.U = U0;
x = prepx(prob, x);
par.prob.mu = 0;
if ~guess_Delta,
    x.Delta0 = max(size(x.V,1), size(x.U,1)) * size(x.V,2) / 8; % As suggested by BV
end
par.x = x;

[X_lrgeom, out_lrgeom] = geom_rtr_common(Matcom, par);
fprintf('Total time is %f \n', X_lrgeom.infos.iter_time(end));



%% Run right invariant GH TR by Mishra et al. 2012
fprintf('----------- Right inv. GH^T TR ---------- \n');

clear x ;
x.G = G0;
x.H = H0;
x.GtG = x.G'*x.G;
x.HtH = x.H'*x.H;
x = prepx_right_inv_gh(prob, x);
if ~guess_Delta,
    x.Delta0 = r;
end

par.x = x;
[X_right_inv_GH, out_right_inv_GH] = geom_rtr_common_right_inv_gh(Matcom,par);
fprintf('Total time is %f \n', X_right_inv_GH.infos.iter_time(end));




%% Run TR by Mishra et al. 2011
fprintf('----------- UBV^T TR ---------- \n');
clear x;
x.U = U0;
x.B = B0;
x.V = V0;
x = prepx_ubv(prob, x);
if ~guess_Delta,
    x.Delta0 = r;
end
par.x = x;


[X_UBV, out_UBV] = geom_rtr_common_ubv(Matcom,par);


%% Run UY trust-region algorithm
fprintf('----------- UY^T TR ---------- \n');

clear x;

x.U = U0;
x.Y = V0*B0;
x.YtY = x.Y'*x.Y;
x = prepx_uy(prob, x);
if ~guess_Delta,
    x.Delta0 = r;
end
par.x = x;

[X_right_inv_UY,out_right_inv_UY] = geom_rtr_common_uy(Matcom,par);






%% Run qGeomMC TR by Mishra et al. 2012
fprintf('----------- qGeomMC TR ---------- \n');

clear x ;
x.G = G0;
x.H = H0;
x.GtG = x.G'*x.G;
x.HtH = x.H'*x.H;
x.invGtG = x.GtG \ eye(size(x.GtG));
x.invHtH = x.HtH \ eye(size(x.HtH));
x = prepx_gh(prob, x);
if ~guess_Delta,
    x.Delta0 = min(d1, d2)*r;
end
par.x = x;

[X_GH, out_GH] = geom_rtr_common_gh(Matcom,par);

fprintf('Total time is %f \n', X_GH.infos.iter_time(end));




%% Run RTRMC by Boumal and Absil 2011,
fprintf('----------- RTRMC ---------- \n');

lambda = 0; % no regularization
problem_rtrmc = buildproblem(Matcom.row', Matcom.col', Matcom.values_Omega', ones(m,1), d1, d2, r, lambda);

problem_rtrmc.A = Matcom.left;
problem_rtrmc.B = Matcom.right';

problem_rtrmc.data_ts.m = prob.data_ts.m;
problem_rtrmc.data_ts.rows = uint32(prob.data_ts.rows);
problem_rtrmc.data_ts.cols = uint32(prob.data_ts.cols);
problem_rtrmc.data_ts.entries = prob.data_ts.entries;

% Setup the options for the RTRMC algorithm
opts_rtrmc.maxouter = max_iter;
opts_rtrmc.maxinner = max_inner;
opts_rtrmc.abstol = tol;
opts_rtrmc.verbosity = verbosity;
opts_rtrmc.order = 2; % RTR
opts_rtrmc.computeRMSE = compute_predictions;

% Call the algorithm
[U_rtrmc W_rtrmc stats_rtrmc] = rtrmc(problem_rtrmc, opts_rtrmc, U0);

% store information
infos_rtrmc.costs = [stats_rtrmc.fx];
infos_rtrmc.iter_time = [stats_rtrmc.iter_time];
infos_rtrmc.numit = cumsum([stats_rtrmc.numinner]);

fprintf('Total time is %f \n', infos_rtrmc.iter_time(end));





%% Plots
% Scaling for plots
X_GH.infos.costs = 2*X_GH.infos.costs/prob.m; % mse
X_lrgeom.infos.costs = 2*X_lrgeom.infos.costs/prob.m; % mse
X_right_inv_GH.infos.costs = 2*X_right_inv_GH.infos.costs/prob.m; % mse
infos_rtrmc.costs = 2*infos_rtrmc.costs / prob.m;
X_UBV.infos.costs = 2*X_UBV.infos.costs/prob.m; % mse
X_right_inv_UY.infos.costs = 2*X_right_inv_UY.infos.costs/prob.m; % mse

% Cost versus iterations
figure;
fs = 20;
semilogy(0 : length(X_GH.infos.costs) - 1, X_GH.infos.costs,'-.x','Color','blue','LineWidth',2,'markerSize',10);
hold on;
semilogy(0 : length(X_lrgeom.infos.costs) - 1, X_lrgeom.infos.costs,'-.x','Color','black','LineWidth',2,'markerSize',10);
semilogy(0 : length(X_right_inv_GH.infos.costs) - 1, X_right_inv_GH.infos.costs,'-.x','Color','red','LineWidth',2,'markerSize',10);
semilogy(0 : length(infos_rtrmc.costs) -1, infos_rtrmc.costs,'-.x','Color',[0.5 0.3 .3],'LineWidth',2,'markerSize',10);
semilogy(0 : length(X_UBV.infos.costs) - 1, X_UBV.infos.costs,'-.x','Color','green','LineWidth',2,'markerSize',10);
semilogy(0 : length(X_right_inv_UY.infos.costs) - 1, X_right_inv_UY.infos.costs,'-.x','Color','cyan','LineWidth',2,'markerSize',10);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Number of iterations'); ylabel('Cost');
legend('qGeomMC TR', 'LRGeom TR', 'Right inv. GH^T TR', 'RTRMC', 'UBV^T TR','UY^T TR');
legend 'boxoff';
box off;
title(['qGeomMC TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling), ', Condition No. ', num2str(condition_number)])


% Cost versus time
figure;
fs = 20;
semilogy(X_GH.infos.iter_time,X_GH.infos.costs,'-.x','Color','blue','LineWidth',2,'markerSize',10);
hold on;
semilogy(X_lrgeom.infos.iter_time,X_lrgeom.infos.costs,'-.x','Color','black','LineWidth',2,'markerSize',10);
semilogy(X_right_inv_GH.infos.iter_time,X_right_inv_GH.infos.costs,'-.x','Color','red','LineWidth',2,'markerSize',10);
semilogy(infos_rtrmc.iter_time, infos_rtrmc.costs,'-.x','Color',[0.5 0.3 .3],'LineWidth',2,'markerSize',10);
semilogy(X_UBV.infos.iter_time,X_UBV.infos.costs,'-.x','Color','green','LineWidth',2,'markerSize',10);
semilogy(X_right_inv_UY.infos.iter_time,X_right_inv_UY.infos.costs,'-.x','Color','cyan','LineWidth',2,'markerSize',10);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Time taken (sec.)'); ylabel('Cost');
legend('qGeomMC TR', 'LRGeom TR', 'Right inv. GH^T TR', 'RTRMC', 'UBV^T TR','UY^T TR');
legend 'boxoff';
box off;
title(['qGeomMC TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling),', Condition No. ', num2str(condition_number)])


% Cost versus inner-iteration#
figure;
fs = 20;
semilogy(X_GH.infos.numit,X_GH.infos.costs,'b--d','LineWidth',2);
hold on;
semilogy(X_lrgeom.infos.numit,X_lrgeom.infos.costs,'k--d','LineWidth',2);
semilogy(X_right_inv_GH.infos.numit,X_right_inv_GH.infos.costs,'r--d','LineWidth',2);
semilogy(infos_rtrmc.numit,infos_rtrmc.costs,'--d','color',[0.5 0.3 .3],'LineWidth',2);
semilogy(X_UBV.infos.numit,X_UBV.infos.costs,'g--d','LineWidth',2);
semilogy(X_right_inv_UY.infos.numit,X_right_inv_UY.infos.costs,'c--d','LineWidth',2);

hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Inner iterations'); ylabel('Cost function');
legend('qGeomMC TR', 'LRGeom TR', 'Right inv. GH^T TR', 'RTRMC', 'UBV^T TR','UY^T TR');
legend 'boxoff';
box off;
title(['qGeomMC TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling),', Condition No. ', num2str(condition_number)])


% Compute test error
if compute_predictions,
    
    infos_rtrmc.test_error = [stats_rtrmc.dist];
    
    X_GH.infos.test_error = 2*X_GH.infos.test_error/prob.data_ts.m; %scaling
    X_lrgeom.infos.test_error = 2*X_lrgeom.infos.test_error/prob.data_ts.m; %scaling
    X_right_inv_GH.infos.test_error = 2*X_right_inv_GH.infos.test_error/prob.data_ts.m; %scaling
    infos_rtrmc.test_error = 2*infos_rtrmc.test_error /prob.data_ts.m;
    X_UBV.infos.test_error = 2*X_UBV.infos.test_error/prob.data_ts.m; %scaling
    X_right_inv_UY.infos.test_error = 2*X_right_inv_UY.infos.test_error/prob.data_ts.m; %scaling
    
    fs = 20;
    figure,
    semilogy(0:length(X_GH.infos.test_error)-1,(X_GH.infos.test_error),'-.x','Color','blue','LineWidth',2,'markerSize',10);
    hold on;
    semilogy(0:length(X_lrgeom.infos.test_error)-1,(X_lrgeom.infos.test_error),'-.x','Color','black','LineWidth',2,'markerSize',10);
    semilogy(0:length(X_right_inv_GH.infos.test_error)-1,(X_right_inv_GH.infos.test_error),'-.x','Color','red','LineWidth',2,'markerSize',10);
    semilogy(0:length(infos_rtrmc.test_error) -1, infos_rtrmc.test_error,'-.x','Color',[0.5 0.3 .3],'LineWidth',2,'markerSize',10);
    semilogy(0:length(X_UBV.infos.test_error)-1,(X_UBV.infos.test_error),'-.x','Color','green','LineWidth',2,'markerSize',10);
    semilogy(0:length(X_right_inv_UY.infos.test_error)-1,(X_right_inv_UY.infos.test_error),'-.x','Color','cyan','LineWidth',2,'markerSize',10);
    hold off;
    ax1 = gca;
    set(ax1,'FontSize',fs);
    xlabel(ax1,'Number of iterations','FontSize',fs);
    ylabel(ax1,'Test error','FontSize',fs);
    legend('qGeomMC TR', 'LRGeom TR', 'Right inv. GH^T TR', 'RTRMC', 'UBV^T TR','UY^T TR');
    legend 'boxoff';
    box off;
    title(['qGeomMC TR: ',num2str(d1),' by ',num2str(d2),', rank ',num2str(true_rank),', OS ', num2str(over_sampling),', Condition No. ', num2str(condition_number)])
end


