function [model, infos] = CG_gh(data_ts, data_ls, model, params)
%[model, infos] = CG_gh(data_ts, data_ls, model, params)
%
%
% Parementers
% -----------
% model: Stuctured array with low-rank factors
%
% data_ls: data_ls.rows: rows of the given entries
%        : data_ls.cols: columns of the given entries
%        : data_ls.entries: values of the given entries
%        : data_ls.nentries: number of given entries
%
% data_ts: data_ts.Gs: Original left low-rank factor
%        : datat_ts.Hs: Original right low-rank factor
%
% params.tol : Tolerance value used in the stopping criterion
% params.verb: Verbosity
% params.sigma_armijo: The constant factor used in the Armijo linesearch
% params.ls_maxiter: Maximum of linesearch allowed
% params.beta_type : 'off' means Gradient descnet, 
%                    'P-R' means Conjugate gradient with Polak-Ribiere+,
%                    'F-R' means Conjugate gradient with Fletcher-Reeves
%                    'H-S' means Conjugate gradient with Hestenes-Stiefel+
% params.exact_linesearch : true for exact linesearch,
%                         : false for adaptive stepsize (arXiv:1209.0430, Equation (13))
% params.compute_predictions: To compute test predictions at each iteration
%                             by randomly choosing a subset of entries  
% 
% Output:
% -------
% model: Structured array with final low-rank factors
% infos: Structured array with additional information
%
%
% This is an algorithmic implementation of
% "A Riemannian geometry for low-rank matrix completion",
% B. Mishra, K. Adithya Apuroop and R. Sepulchre,
% Technical report (arXiv:1211.1550), 2012.
% 
% Author:
% Bamdev Mishra, <b.mishra@ulg.ac.be>, 2012


% Set default parameters
t_begin = tic();

params = default_params(params);
tol = params.tol;
verb = params.verb;
maxiter = params.maxiter;
max_step = params.max_step;
ls_maxiter = params.ls_maxiter;

% Problem dimensions
ORTH_VALUE = params.orth_value; % Measure of orthogoanlity used in the CG schemes
r = size(model.G, 2);
d1 = size(model.G, 1);
d2 = size(model.H, 1);

% Compute cost function
n = data_ls.nentries;
preds_ls = partXY(model.G', model.H', data_ls.rows, data_ls.cols, n)';
errors = (preds_ls - data_ls.entries);
M = model.M; % Euclidean gradient
updateSparse(M, errors); % Sparse update of gradient in d1 \times d2 Euclidena space
loss = (errors'*errors)/2;%   0.5*(errors.^2);


% If interested in computing recovery
if params.compute_predictions,
    preds_test = partXY(model.G', model.H',data_ts.rows,data_ts.cols, data_ts.nentries)';
    errors_test = preds_test - data_ts.entries;
    cost_test = (errors_test'*errors_test) / data_ts.nentries;
    infos.test_error =  cost_test;    
end

cost = loss; 
invHtH = eye(r) / model.HtH;
invGtG = eye(r) / model.GtG;
model.HinvHtH = model.H * invHtH;
model.GinvGtG = model.G * invGtG;
grad.G = M*model.HinvHtH;
grad.H = M'*model.GinvGtG;

% grad.G = M*model.H;
% grad.H = M'*model.G;
% grad.G = grad.G * invHtH;
% grad.H = grad.H * invGtG;

ip_grad_grad = inner_product_gh(model, grad, grad);


if verb,
    fprintf('[%0.4d] Cost = %5.3e\n', 0, (2/n)*cost); % mean least square loss
end


% Information
infos.sqnormgrad = [];
infos.beta = 0;
infos.linesearch = [];
infos.costs = (2/n)*cost; % mean least square loss
infos.iter_time = 0;

% Local variables
j = 0; % Linesearch along the search direction
dir.G = -grad.G;
dir.H = -grad.H;
ip_grad_dir = -ip_grad_grad;
linesearch_fail = 0;


for iter = 1 : maxiter
    
    % Compute the first step.
    % If we don't do an exact line search all the time,
    % we still need it once for the starting stepsize guess.
    if params.exact_linesearch || iter==1
        [alpha, cost_new, errors]  = guess_linesearch_gh(model, errors, dir, data_ls);
        % No sign correction needed, guess_linesearch gives the corrected sign
        max_step = alpha;
        model_new = model;
        Gi = model.G;
        Hi = model.H;
       
        model_new.G = Gi + alpha * dir.G;
        model_new.H = Hi + alpha * dir.H;
        sigma_armijo = 1e-4; % Keep a low value or 0 for an exact search
        
    else % Adaptive stepsize as proposed by MMBS in arXiv:1209.0430
        if j == 0,
            max_step = 2*max_step;
        elseif j > 1,
            max_step = 2*alpha;
        end
        alpha = max_step;
        alpha = -sign(ip_grad_dir)*abs(alpha); % Sign correction
    
        model_new = model;
        Gi = model.G;
        Hi = model.H;
        
        model_new.G = Gi + alpha * dir.G;
        model_new.H = Hi + alpha * dir.H;
        
        preds_ls = partXY(model_new.G', model_new.H', data_ls.rows, data_ls.cols, n)';
        errors = (preds_ls - data_ls.entries);
        cost_new = (errors'*errors)/2;
        sigma_armijo = params.sigma_armijo; % User specified value
    end
       
    armijo = (cost - cost_new) >= sigma_armijo * abs(alpha * ip_grad_dir);
    j = 0;
    while ~armijo,
        j = j + 1;
        alpha = alpha / 2;
        model_new.G = Gi + alpha * dir.G;
        model_new.H = Hi + alpha * dir.H;
        preds_ls = partXY(model_new.G', model_new.H', data_ls.rows, data_ls.cols, n)';
        errors = (preds_ls - data_ls.entries);
        cost_new = (errors'*errors)/2;
        armijo = (cost - cost_new) >= sigma_armijo * abs(alpha * ip_grad_dir);
        if j > ls_maxiter,
            fprintf('*** Linesearch failed. Quitting...\n ');
            linesearch_fail = 1;
            break;
        end
    end
    model_new.GtG = model_new.G'*model_new.G;
    model_new.HtH = model_new.H'*model_new.H;
%     model_new = quotient_gh_balancing_update(model_new); % Numerical conditioning
    
    updateSparse(M, errors);
    
    invHtH_new = eye(r) / model_new.HtH;
    invGtG_new = eye(r) / model_new.GtG;
    model_new.HinvHtH = model_new.H * invHtH_new;
    model_new.GinvGtG = model_new.G * invGtG_new;
    grad_new.G = M*model_new.HinvHtH;
    grad_new.H = M'*model_new.GinvGtG;

%     grad_new.G = M*model_new.H;
%     grad_new.H = M'*model_new.G;
%     grad_new.G = grad_new.G * invHtH_new;
%     grad_new.H = grad_new.H * invGtG_new;
    
    
    
    ip_grad_grad_new = inner_product_gh(model_new, grad_new, grad_new);
   
    % Collect information  
    infos.sqnormgrad = [infos.sqnormgrad; ip_grad_grad_new];
    infos.linesearch = [infos.linesearch; j];
    infos.costs = [infos.costs; (2/n)*cost_new];
    infos.iter_time = [infos.iter_time; toc(t_begin)];
    
    % If interested in computing recovery
    if params.compute_predictions,
        preds_test = partXY(model_new.G', model_new.H',data_ts.rows,data_ts.cols, data_ts.nentries)';
        errors_test = preds_test - data_ts.entries;
        cost_test = (errors_test'*errors_test) / data_ts.nentries;
        infos.test_error =[infos.test_error; cost_test];
    end
    
    % Print
    if verb,
        fprintf('[%0.4d] Cost = %7.3e, #extra linesearch = %i, Gradient norm sq. = %7.3e, Step = %7.3e \n', iter, (2/n)*cost_new, j, ip_grad_grad_new, alpha);
    end
    
    % Stopping criterion
    if (2/n)*cost_new < tol || linesearch_fail,
        break;
    end
        
    % Algorithm type
    if strcmpi(params.beta_type, 'off') % Gradient Descent
        beta = 0;
        dir.G = -grad_new.G;
        dir.H = -grad_new.H;        
    else
        grad_old = vector_transport_gh(model, grad, model_new); % vector transport to the new point
        orth_grads = inner_product_gh(model_new, grad_old, grad_new)/ip_grad_grad_new;
        
        if abs(orth_grads) >= ORTH_VALUE
            beta = 0;
            dir.G = -grad_new.G;
            dir.H = -grad_new.H;
            
        else % Compute the CG modification
            dir = vector_transport_gh(model, dir, model_new); % vector transport to the new point
            
            if strcmp(params.beta_type, 'F-R')  % Fletcher-Reeves
                % vector grad(new) - transported grad(current)
                beta = ip_grad_grad_new / ip_grad_grad;
                
            elseif strcmp(params.beta_type, 'P-R')  % Polak-Ribiere+
                % vector grad(new) - transported grad(current)
                diff.G = grad_new.G - grad_old.G;
                diff.H = grad_new.H - grad_old.H;
                ip_diff = inner_product_gh(model_new, grad_new, diff);
                beta = ip_diff / ip_grad_grad;
                beta = max(0,beta);
                
            elseif strcmp(params.beta_type, 'H-S')  % Hestenes-Stiefel+
                diff.G = grad_new.G - grad_old.G;
                diff.H = grad_new.H - grad_old.H;
                ip_diff = inner_product_gh(model_new, grad_new, diff);
                ip_diff_dir = inner_product_gh(model_new, diff, dir);
                beta = ip_diff / ip_diff_dir;
                beta = max(0, beta);
            else
                error(['Unknown params.beta_type. ' ...
                       'Should be off, F-R, P-R or H-S.']);
            end
            
            dir.G = -grad_new.G + beta*dir.G;
            dir.H = -grad_new.H + beta*dir.H;            
        end
        
    end
    
    ip_grad_dir_new = inner_product_gh(model_new, dir, grad_new);
    
    % If CG update produces an accent direction, we either reverse it or
    % switch to the negative gradient direction.
    % You came to the wrong neighborhood, bro!
    if sign(ip_grad_dir_new) > 0,
%         % Either we reverse it.
%         ip_grad_dir_new = -ip_grad_dir_new;
%         dir.G = -dir.G;
%         dir.H = -dir.H;
%         if verb,
%             fprintf('CG warning: An accent direction obtained. Reversed it.\n');
%         end
        
        % Or we switch to the negative gradient direction.
        dir.G = -grad_new.G;
        dir.H = -grad_new.H;
        beta = 0;
        ip_grad_dir_new = -ip_grad_grad_new;
        if verb,
            fprintf('CG warning: An accent direction obtained. Switched to negative gradient dir.\n');
        end

    end
    
    % Note: By this time we should have properly chosen the CG direction.
    % Next: We collect the information and prepare for the next iteration.
       
    infos.beta = [infos.beta; beta];
    
    % New iterate
    grad = grad_new;
    ip_grad_grad = ip_grad_grad_new;
    ip_grad_dir = ip_grad_dir_new;
    model = model_new;
    cost = cost_new;
end

model = model_new;

return;
